use chrono::{NaiveDate, NaiveDateTime};
use sqlx::{query, PgPool};

use super::{event_changeset::EventChangeset, Clothing, Location, Result};

#[derive(Clone, Debug)]
pub struct Event {
    pub id: i32,
    pub start: NaiveDateTime,
    pub end: NaiveDateTime,
    pub name: String,
    pub location_id: i32,
    pub location: Option<Location>,
    pub voluntary_wachhabender: bool,
    pub voluntary_fuehrungsassistent: bool,
    pub amount_of_posten: i16,
    pub clothing: Clothing,
    pub canceled: bool,
    pub note: Option<String>,
}

impl Event {
    pub async fn create(pool: &PgPool, changeset: EventChangeset) -> Result<()> {
        query!(r#"
            INSERT INTO event (startTimestamp, endTimestamp, name, locationId, voluntaryWachhabender, voluntaryFuehrungsassistent, amountOfPosten, clothing, note)
            VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9);
            "#,
            changeset.time.0.and_utc(), changeset.time.1.and_utc(), changeset.name, changeset.location_id, changeset.voluntary_wachhabender, changeset.voluntary_fuehrungsassistent, changeset.amount_of_posten, changeset.clothing, changeset.note).execute(pool).await?;

        Ok(())
    }

    pub async fn read_all_by_date_and_area_including_location(
        pool: &PgPool,
        date: NaiveDate,
        area_id: i32,
    ) -> Result<Vec<Event>> {
        let records = query!(
            r#"
            SELECT
                event.id AS eventId,
                event.startTimestamp,
                event.endTimestamp,
                event.name,
                event.locationId,
                event.voluntaryWachhabender,
                event.voluntaryFuehrungsassistent,
                event.amountOfPosten,
                event.clothing,
                event.canceled,
                event.note,
                location.id,
                location.name AS locationName,
                location.areaId AS locationAreaId,
                clothing.id AS clothingId,
                clothing.name AS clothingName
            FROM event
            JOIN location ON event.locationId = location.id
            JOIN clothing ON event.clothing = clothing.id
            WHERE starttimestamp::date = $1
            AND location.areaId = $2;
            "#,
            date,
            area_id
        )
        .fetch_all(pool)
        .await?;

        let events = records
            .into_iter()
            .map(|record| Event {
                id: record.eventid,
                start: record.starttimestamp.naive_utc(),
                end: record.endtimestamp.naive_utc(),
                name: record.name.to_string(),
                location_id: record.locationid,
                location: Some(Location {
                    id: record.locationid,
                    name: record.locationname.to_string(),
                    area_id: record.locationareaid,
                    area: None,
                }),
                voluntary_wachhabender: record.voluntarywachhabender,
                voluntary_fuehrungsassistent: record.voluntaryfuehrungsassistent,
                amount_of_posten: record.amountofposten,
                clothing: Clothing {
                    id: record.clothingid,
                    name: record.clothingname,
                },
                canceled: record.canceled,
                note: record.note,
            })
            .collect();

        Ok(events)
    }

    pub async fn read_by_id_including_location(pool: &PgPool, id: i32) -> Result<Option<Event>> {
        let record = query!(
            r#"
            SELECT
                event.id AS eventId,
                event.startTimestamp,
                event.endTimestamp,
                event.name,
                event.locationId,
                event.voluntaryWachhabender,
                event.voluntaryFuehrungsassistent,
                event.amountOfPosten,
                event.clothing,
                event.canceled,
                event.note,
                location.id,
                location.name AS locationName,
                location.areaId AS locationAreaId,
                clothing.id AS clothingId,
                clothing.name AS clothingName
            FROM event
            JOIN location ON event.locationId = location.id
            JOIN clothing ON event.clothing = clothing.id
            WHERE event.id = $1;
            "#,
            id
        )
        .fetch_optional(pool)
        .await?;

        let event = record.map(|record| Event {
            id: record.eventid,
            start: record.starttimestamp.naive_utc(),
            end: record.endtimestamp.naive_utc(),
            name: record.name.to_string(),
            location_id: record.locationid,
            location: Some(Location {
                id: record.locationid,
                name: record.locationname.to_string(),
                area_id: record.locationareaid,
                area: None,
            }),
            voluntary_wachhabender: record.voluntarywachhabender,
            voluntary_fuehrungsassistent: record.voluntaryfuehrungsassistent,
            amount_of_posten: record.amountofposten,
            clothing: Clothing {
                id: record.clothingid,
                name: record.clothingname,
            },
            canceled: record.canceled,
            note: record.note,
        });

        Ok(event)
    }

    pub async fn update(pool: &PgPool, id: i32, changeset: EventChangeset) -> Result<()> {
        query!(r#"
            UPDATE event SET startTimestamp = $1, endTimestamp = $2, name = $3, locationId = $4, voluntaryWachhabender = $5, voluntaryFuehrungsassistent = $6, amountOfPosten = $7, clothing = $8, note = $9 WHERE id = $10;
            "#,
            changeset.time.0.and_utc(),
            changeset.time.1.and_utc(),
            changeset.name,
            changeset.location_id,
            changeset.voluntary_wachhabender,
            changeset.voluntary_fuehrungsassistent,
            changeset.amount_of_posten,
            changeset.clothing,
            changeset.note, id)
            .execute(pool)
            .await?;

        Ok(())
    }

    pub async fn update_cancelation(pool: &PgPool, id: i32, canceled: bool) -> Result<()> {
        query!(
            "UPDATE event SET canceled = $1 WHERE id = $2;",
            canceled,
            id
        )
        .execute(pool)
        .await?;

        Ok(())
    }

    pub async fn delete(pool: &PgPool, id: i32) -> Result<()> {
        query!("DELETE FROM event WHERE id = $1;", id)
            .execute(pool)
            .await?;

        Ok(())
    }
}
